#include <sta/math/algorithms/kalmanFilter.hpp>
#include <sta/math/linalg/linalg.hpp>
#include <sta/debug/debug.hpp>
#include <sta/debug/assert.hpp>

namespace sta
{

namespace math
{

KalmanFilter::KalmanFilter(matrix F, matrix B, matrix H, matrix Q, matrix R) : F_{F}, B_{B}, H_{H}, Q_{Q}, R_{R}, n_{F.get_cols()}
{
	STA_ASSERT_MSG(F.get_rows() == B.get_rows(), "#rows mismatch: F, B!");
	STA_ASSERT_MSG(F.get_cols() == H.get_cols(), "#cols mismatch: F, H!");
	STA_ASSERT_MSG(F.get_rows() == Q.get_rows(), "#rows mismatch: F, Q!");
	STA_ASSERT_MSG(Q.get_cols() == Q.get_rows(), "Q not square!");
    STA_ASSERT_MSG(F.get_cols() == F.get_rows(), "F not square!");
	STA_ASSERT_MSG(H.get_rows() == R.get_rows(), "#rows mismatch: H, R");
	STA_ASSERT_MSG(R.get_rows() == R.get_cols(), "#R not square");
    identity_ = matrix::eye(n_);
}

KalmanFilter::~KalmanFilter()
{
    // Destructor implementation
}

KalmanState KalmanFilter::predict(KalmanState state, matrix u)
{
    // Predict step implementation
    // Update the state based on the system dynamics
    state.x = F_ * state.x + B_ * u;
    // Update the error covariance matrix
    state.error = F_ * state.error * F_.T() + Q_;
    return state;
}

KalmanState KalmanFilter::correct(KalmanState state, matrix z)
{
    // Correct step implementation
    // Calculate the Kalman gain
    matrix K = state.error * H_.T() * linalg::inv(H_ * state.error * H_.T() + R_);
    K.show_serial();
    // Update the state based on the measurement
    state.x = state.x + K * (z - H_ * state.x); //TODO check transpose
    state.x.show_serial();
    // Update the error covariance matrix
    state.error = (identity_ - K * H_) * state.error;
    return state;
}

KalmanState KalmanFilter::correct(KalmanState state, matrix z, matrix H, matrix R)
{
    // Correct step implementation
    // Calculate the Kalman gain
    matrix K = state.error * H.T() * linalg::inv(H * state.error * H.T() + R);
    K.show_serial();
    // Update the state based on the measurement
    state.x = state.x + K * (z - H * state.x); //TODO check transpose
    state.x.show_serial();
    // Update the error covariance matrix
    state.error = (matrix::eye(state.x.get_rows()) - K * H) * state.error;
    return state;
}

} // namespace math

} // namespace sta
